from bcc import BPF

# eBPF Program
#########################################
bpf_text = """ 
#include <net/inet_sock.h>
#include <linux/socket.h>
#include <bcc/proto.h>

struct packet_evt_t {
    u64 test;
};
BPF_PERF_OUTPUT(packet_evt);


// separate data structs for ipv4 and ipv6
struct ipv4_data_t {
    // XXX: switch some to u32's when supported
    u64 type;
    u64 pid;
    u64 ip;
    u64 dlen;
    u64 saddr;
    u64 daddr;
    u64 lport;
    u64 dport;

};
BPF_PERF_OUTPUT(ipv4_events);

struct ipv6_data_t {
    u64 type;
    u64 pid;
    u64 ip;
    u64 dlen;
    unsigned __int128 saddr;
    unsigned __int128 daddr;
    u64 lport;
    u64 dport;

};
BPF_PERF_OUTPUT(ipv6_events);


static int do_packet_event(struct pt_regs *ctx, u64 typ) {
    struct packet_evt_t evt = {
       .test = 42,
    };
    evt.test = typ;
    packet_evt.perf_submit(ctx, &evt, sizeof(evt));
    return 1;
}


int kprobe__sock_sendmsg(struct pt_regs *ctx, struct socket *sock, struct msghdr *msg)
{
    struct ipv4_data_t data4 = {.pid = 0, .ip = 4, .type = 42, .dlen = 0 };
    struct ipv6_data_t data6 = {.pid = 0, .ip = 6, .type = 42, .dlen = 0 };
    u8 *cursor = 0;
    struct sock *sk = sock->sk;
    struct sock *skp = NULL;
    bpf_probe_read(&skp, sizeof(skp), &sk);

    u16 family = 0;
    u16 lport = 0, dport = 0;
    u32 pid = bpf_get_current_pid_tgid();


    bpf_probe_read(&family, sizeof(family), &skp->__sk_common.skc_family);
    bpf_probe_read(&lport, sizeof(lport), &skp->__sk_common.skc_num);
    bpf_probe_read(&dport, sizeof(dport), &skp->__sk_common.skc_dport);
    dport = ntohs(dport);
    data4.pid = data6.pid = pid;

    if (pid == 2414) {
      return 1;
    }


    struct packet_evt_t evt = { .test = 42, };
    // do_packet_event(ctx, ip->nextp);

    evt.test = 43;
    // data4.dlen = data6.dlen = msg->msg_iter.count;
    struct iovec *iov = (void *) msg->msg_iter.iov; 

    u32 test = (u32) (msg->msg_iter.iov);

    data4.dlen = data6.dlen = iov->iov_len; 

    if (family == AF_INET) {
        bpf_probe_read(&data4.saddr, sizeof(u32),
            &skp->__sk_common.skc_rcv_saddr);
        bpf_probe_read(&data4.daddr, sizeof(u32),
            &skp->__sk_common.skc_daddr);
        data4.lport = lport;
        data4.dport = dport;

        ipv4_events.perf_submit(ctx, &data4, sizeof(data4));
    } else if (family == AF_INET6) {
        bpf_probe_read(&data6.saddr, sizeof(data6.saddr),
            &skp->__sk_common.skc_v6_rcv_saddr.in6_u.u6_addr32);
        bpf_probe_read(&data6.daddr, sizeof(data6.daddr),
            &skp->__sk_common.skc_v6_daddr.in6_u.u6_addr32);
        data6.lport = lport;
        data6.dport = dport;

        ipv6_events.perf_submit(ctx, &data6, sizeof(data6));
    } else if (family == AF_UNIX) {
      return 1;
    }
 
    // packet_evt.perf_submit(ctx, &evt, sizeof(evt));

    return 1;
}

int kprobe__ipv6_rcv(struct pt_regs *ctx, struct sk_buff *skb, struct net_device *dev, struct packet_type *pt, struct net_device *orig_dev)
{
  u32 data = 0;
  // const struct iphdr *iph = ip_hdr(skb);
  bpf_probe_read(&data, sizeof(u32), skb->data);
  do_packet_event(ctx, data);
  return 1;
}

int kretprobe__ipv6_rcv(struct pt_regs *ctx, struct sk_buff *skb, struct net_device *dev, struct packet_type *pt, struct net_device *orig_dev)
{
  u32 data = 255;
  do_packet_event(ctx, data);
  return 1;
}


"""
#########################################

# We need ctypes to parse the event structure
import ctypes
import ctypes as ct
from socket import *
from time import strftime
from struct import pack

class PacketEvt(ctypes.Structure):
    _fields_ = [
        ("test",   ctypes.c_ulonglong),
    ]



# event data
class Data_ipv4(ct.Structure):
    _fields_ = [
        ("type", ct.c_ulonglong),
        ("pid", ct.c_ulonglong),
        ("ip", ct.c_ulonglong),
        ("dlen", ct.c_ulonglong),
        ("saddr", ct.c_ulonglong),
        ("daddr", ct.c_ulonglong),
        ("lport", ct.c_ulonglong),
        ("dport", ct.c_ulonglong),
    ]

class Data_ipv6(ct.Structure):
    _fields_ = [
        ("type", ct.c_ulonglong),
        ("pid", ct.c_ulonglong),
        ("ip", ct.c_ulonglong),
        ("dlen", ct.c_ulonglong),
        ("saddr", (ct.c_ulonglong * 2)),
        ("daddr", (ct.c_ulonglong * 2)),
        ("lport", ct.c_ulonglong),
        ("dport", ct.c_ulonglong),
    ]



# Declare event printer
def print_pkt_event(cpu, data, size):
    event = ctypes.cast(data, ctypes.POINTER(PacketEvt)).contents
    print("Packet event with test data %x" % (
        event.test,
    ))

# process event
def print_ipv4_event(cpu, data, size):
    event = ct.cast(data, ct.POINTER(Data_ipv4)).contents
    atime = strftime("%H:%M:%S")
    print("%-8s %-6d %-2d %-20s %1s> %-20s %d" % (
        atime, event.pid, event.ip,
        "%s:%d" % (inet_ntop(AF_INET, pack('I', event.saddr)), event.lport),
        event.type,
        "%s:%s" % (inet_ntop(AF_INET, pack('I', event.daddr)), event.dport), event.dlen))

def print_ipv6_event(cpu, data, size):
    event = ct.cast(data, ct.POINTER(Data_ipv6)).contents
    print("%-8s %-6d %-2d %-20s %1s> %-20s %d" % (
        strftime("%H:%M:%S"), event.pid, event.ip,
        "%s:%d" % (inet_ntop(AF_INET6, event.saddr), event.lport),
        event.type,
        "%s:%d" % (inet_ntop(AF_INET6, event.daddr), event.dport), event.dlen))





# 2. Build and Inject program
b = BPF(text=bpf_text)

# 3. Print debug output
# while True:
#    print b.trace_readline()
b["packet_evt"].open_perf_buffer(print_pkt_event)
b["ipv4_events"].open_perf_buffer(print_ipv4_event)
b["ipv6_events"].open_perf_buffer(print_ipv6_event)


while True:
    b.kprobe_poll()






